from psycopg2.pool import AbstractConnectionPool, PoolError
from psycopg2 import extensions as _ext
import threading
import time


class ConnectionPool(AbstractConnectionPool):
    def __init__(
        self,
        idle_time: int = 2,
        *args,
        **kwargs,
    ):
        AbstractConnectionPool.__init__(self, *args, **kwargs)
        # Construct Semaphore with maxconn. This will enable restricting
        # the number of connections that are being used to not exceed the
        # number of connections which causes the psycopg2 pool to raise an
        # exception
        self._connectionLock = threading.Semaphore(int(self.maxconn))
        self._idle_pool = []
        self._last_used = {}  # id(conn): time
        self._in_use_connections = {}  # id(conn): conn
        self._lock = threading.Lock()
        # The daemon will run periodically and call putconn for all connections
        # that haven't been used for some amount of time
        self.daemon = threading.Thread(
            target=self.trim_pool, args=(idle_time,), daemon=True
        )
        self.daemon.start()

    def trim_pool(self, number_of_sec: int):
        while True and not self.closed:
            # Yield the cpu to other threads for `number_of_sec` sec.
            # This is the least overhead but a connection might live for almost
            # 2 * number_of_sec sec
            time.sleep(number_of_sec)
            if self.closed:
                # Stops the deamon if the pool has been closed
                break

            with self._lock:
                # This only closes connections that have been idle for
                # `number_of_sec` sec
                t = time.time() - number_of_sec
                keep_conn = []
                close_conn = []
                for conn in self._idle_pool:
                    if self._last_used[id(conn)] > t:
                        keep_conn.append(conn)
                    else:
                        close_conn.append(conn)

                for conn in close_conn:
                    del self._last_used[id(conn)]
                    self._putconn(conn)

                self._idle_pool = keep_conn

    def getconn(self):
        if self.closed:
            raise PoolError("connection pool is closed")
        self._connectionLock.acquire()
        with self._lock:
            if self._idle_pool:
                # Get the oldest conn in the pool
                conn = self._idle_pool.pop(0)
            else:
                # If there is no connection in the pool then we open a new
                # connection
                conn = self._getconn()

            self._in_use_connections[id(conn)] = conn

            return conn

    def putconn(self, conn):
        try:
            if self.closed:
                raise PoolError("connection pool is closed")
            with self._lock:
                # We need to handle all the clean up for the connection
                key = id(conn)
                if not conn.closed:

                    status = conn.info.transaction_status
                    if status == _ext.TRANSACTION_STATUS_UNKNOWN:
                        # server connection lost
                        conn.close()
                    else:
                        if status != _ext.TRANSACTION_STATUS_IDLE:
                            # connection in error or in transaction
                            conn.rollback()
                        self._last_used[key] = time.time()
                        self._idle_pool.append(conn)

                if not self.closed and key in self._in_use_connections:
                    del self._in_use_connections[key]
        finally:
            self._connectionLock.release()

    def closeall(self):
        with self._lock:
            self._closeall()



from functools import wraps
from .cursor import Cursor


def pooled_cursor(m):
    """
    Decorator that assumes that wrapped method accepts cursor as parameter.
    If there is `DictCursor` in the method request then this
    function gets a connection from connection pool. It creates a new cursor, delegates
    to the method, then commits, closes the cursors and return connection to pool.

    If a parameter of the method is a `DictCursor` this wrapper does nothing
    and expects the calling method to handle commiting, and closing the cursor,
    and returning the connection to the pool.
    :param method to decorate with pooled cursor
    :return:
    """

    @wraps(m)
    def wrapper(self, *args, **kwargs):
        con = None
        cur = None
        result = None
        # Check if there is a DictCursor in the arguments
        if any([isinstance(a, Cursor) for a in args]) or "cursor" in kwargs:
            return m(self, *args, **kwargs)
        try:
            retry_count = 0
            while 1:
                try:
                    con = self.pool.getconn()
                    cur = con.cursor(cursor_factory=Cursor)
                    result = m(self, *args, cursor=cur, **kwargs)
                except Exception as e:
                    if retry_count == 3:
                        raise e
                    retry_count += 1
                    if con:
                        self.pool.putconn(con)
                    continue
                break
            if con:
                con.commit()
            if cur:
                cur.close()
        finally:
            if con:
                self.pool.putconn(con)
        return result

    return wrapper



import psycopg2
import psycopg2.extras
import logging
import time
import os

log_metrics = os.environ.get("METRIC_LOGGING", "FALSE").upper() == "TRUE"


class Cursor(psycopg2.extras.DictCursor):
    """
    Cursor that adds debug and metric logging.

    - If LogLevel is at least DEBUG, then the sql query that is about to be executed
    will be logged
    - If LogLevel is at least INFO and METRIC_LOGGING is TRUE, then the query action,
    table name, and execution time will be logged.
    """

    def execute(self, sql, args=None):
        logger = logging.getLogger(__name__)
        if logger.isEnabledFor(logging.DEBUG):
            # mogrify constructs the sql query that would be sent to the backend
            logger.debug(self.mogrify(sql, args))

        try:
            if logger.isEnabledFor(logging.INFO) and log_metrics:
                t = time.time()

            super().execute(sql, args)

            if logger.isEnabledFor(logging.INFO) and log_metrics and t:
                # Sometimes `sql` is bytes rather than string.
                sql_str = str(sql, "utf-8") if isinstance(sql, bytes) else sql
                # Remove all extra spaces and new line chars
                sql_ = " ".join(sql_str.strip().split()).split()
                if sql_:
                    action = sql_[0].upper()
                    if action == "UPDATE":
                        # UPDATE cases SET ...
                        # For update the table name is the second value in the list
                        table_name = sql_[1]
                    elif action == "INSERT" or action == "DELETE":
                        # INSERT INTO cases ...
                        # DELETE FROM cases ...
                        # For insert and delete the table name is the third value in the
                        # list
                        table_name = sql_[2]
                    elif action == "SELECT":
                        # For select the table name is the value that comes after FROM
                        table_name = sql_[sql_.index("FROM") + 1]
                    else:
                        # We have some queries that start with `WITH ...`, we need to
                        # map those
                        table_name = "UNKNOWN"
                    logger.info(
                        f"ACTION={action} | TABLE={table_name} | TIME={time.time() - t}"
                    )
        except Exception as exc:
            logger.exception("%s: %s" % (exc.__class__.__name__, exc))
            raise
